@@ -1,5 +1,5 @@
 /* server.c --- SASL CRAM-MD5 server side functions.
- * Copyright (C) 2009-2012 Simon Josefsson
+ * Copyright (C) 2009-2020 Simon Josefsson
  *
  * This file is part of GNU SASL Library.
  *
@@ -21,7 +21,7 @@
  */
 
 #ifdef HAVE_CONFIG_H
-# include "config.h"
+#include "config.h"
 #endif
 
 /* Get specification. */
@@ -44,22 +44,24 @@
 #include "printer.h"
 #include "gc.h"
 #include "memxor.h"
+#include "mechtools.h"
 
 #define DEFAULT_SALT_BYTES 12
 #define SNONCE_ENTROPY_BYTES 18
 
 struct scram_server_state
 {
-  int plus;
+  bool plus;
+  Gsasl_hash hash;
   int step;
   char *cbind;
-  char *gs2header; /* copy of client first gs2-header */
-  char *cfmb_str; /* copy of client first message bare */
-  char *sf_str; /* copy of server first message */
+  char *gs2header;		/* copy of client first gs2-header */
+  char *cfmb_str;		/* copy of client first message bare */
+  char *sf_str;			/* copy of server first message */
   char *snonce;
   char *clientproof;
-  char *storedkey;
-  char *serverkey;
+  char storedkey[GSASL_HASH_MAX_SIZE];
+  char serverkey[GSASL_HASH_MAX_SIZE];
   char *authmessage;
   char *cbtlsunique;
   size_t cbtlsuniquelen;
@@ -70,7 +72,8 @@ struct scram_server_state
 };
 
 static int
-scram_start (Gsasl_session * sctx, void **mech_data, int plus)
+scram_start (Gsasl_session * sctx, void **mech_data,
+	     bool plus, Gsasl_hash hash)
 {
   struct scram_server_state *state;
   char buf[MAX (SNONCE_ENTROPY_BYTES, DEFAULT_SALT_BYTES)];
@@ -82,13 +85,19 @@ scram_start (Gsasl_session * sctx, void **mech_data, int plus)
     return GSASL_MALLOC_ERROR;
 
   state->plus = plus;
+  state->hash = hash;
 
   rc = gsasl_nonce (buf, SNONCE_ENTROPY_BYTES);
   if (rc != GSASL_OK)
     goto end;
 
-  rc = gsasl_base64_to (buf, SNONCE_ENTROPY_BYTES,
-			&state->snonce, NULL);
+  rc = gsasl_base64_to (buf, SNONCE_ENTROPY_BYTES, &state->snonce, NULL);
+#if SCRAMDEBUG
+  if (hash == GSASL_HASH_SHA1)
+    state->snonce = strdup ("3rfcNHYJY1ZVvWVs7j");	// SCRAM-SHA1
+  else
+    state->snonce = strdup ("%hvYDpWUa2RaTCAfuxFIlj)hNlF$k0");	// SCRAM-SHA256
+#endif
   if (rc != GSASL_OK)
     goto end;
 
@@ -96,8 +105,7 @@ scram_start (Gsasl_session * sctx, void **mech_data, int plus)
   if (rc != GSASL_OK)
     goto end;
 
-  rc = gsasl_base64_to (buf, DEFAULT_SALT_BYTES,
-			&state->sf.salt, NULL);
+  rc = gsasl_base64_to (buf, DEFAULT_SALT_BYTES, &state->sf.salt, NULL);
   if (rc != GSASL_OK)
     goto end;
 
@@ -119,31 +127,72 @@ scram_start (Gsasl_session * sctx, void **mech_data, int plus)
 
   return GSASL_OK;
 
- end:
+end:
   free (state->sf.salt);
   free (state->snonce);
   free (state);
   return rc;
 }
 
+#ifdef USE_SCRAM_SHA1
 int
 _gsasl_scram_sha1_server_start (Gsasl_session * sctx, void **mech_data)
 {
-  return scram_start (sctx, mech_data, 0);
+  return scram_start (sctx, mech_data, false, GSASL_HASH_SHA1);
 }
 
 int
 _gsasl_scram_sha1_plus_server_start (Gsasl_session * sctx, void **mech_data)
 {
-  return scram_start (sctx, mech_data, 1);
+  return scram_start (sctx, mech_data, true, GSASL_HASH_SHA1);
+}
+#endif
+
+#ifdef USE_SCRAM_SHA256
+int
+_gsasl_scram_sha256_server_start (Gsasl_session * sctx, void **mech_data)
+{
+  return scram_start (sctx, mech_data, false, GSASL_HASH_SHA256);
 }
 
 int
-_gsasl_scram_sha1_server_step (Gsasl_session * sctx,
-			       void *mech_data,
-			       const char *input,
-			       size_t input_len,
-			       char **output, size_t * output_len)
+_gsasl_scram_sha256_plus_server_start (Gsasl_session * sctx, void **mech_data)
+{
+  return scram_start (sctx, mech_data, true, GSASL_HASH_SHA256);
+}
+#endif
+
+static int
+extract_serverkey (Gsasl_session * sctx,
+		   struct scram_server_state *state,
+		   const char *b64, char *buf)
+{
+  char *bin;
+  size_t binlen;
+  int rc;
+
+  rc = gsasl_base64_from (b64, strlen (b64), &bin, &binlen);
+  if (rc != GSASL_OK)
+    return rc;
+
+  if (binlen != gsasl_hash_length (state->hash))
+    {
+      free (bin);
+      return GSASL_AUTHENTICATION_ERROR;
+    }
+
+  memcpy (buf, bin, binlen);
+
+  free (bin);
+
+  return GSASL_OK;
+}
+
+int
+_gsasl_scram_server_step (Gsasl_session * sctx,
+			  void *mech_data,
+			  const char *input,
+			  size_t input_len, char **output, size_t *output_len)
 {
   struct scram_server_state *state = mech_data;
   int res = GSASL_MECHANISM_CALLED_TOO_MANY_TIMES;
@@ -169,9 +218,8 @@ _gsasl_scram_sha1_server_step (Gsasl_session * sctx,
 	/* In non-PLUS mode, but where have channel bindings data (and
 	   thus advertised PLUS) we reject a client 'y' cbflag. */
 	if (!state->plus
-	    && state->cbtlsuniquelen > 0
-	    && state->cf.cbflag == 'y')
-	    return GSASL_AUTHENTICATION_ERROR;
+	    && state->cbtlsuniquelen > 0 && state->cf.cbflag == 'y')
+	  return GSASL_AUTHENTICATION_ERROR;
 
 	/* Check that username doesn't fail SASLprep. */
 	{
@@ -212,15 +260,15 @@ _gsasl_scram_sha1_server_step (Gsasl_session * sctx,
 	/* Create new nonce. */
 	{
 	  size_t cnlen = strlen (state->cf.client_nonce);
+	  size_t snlen = strlen (state->snonce);
 
-	  state->sf.nonce = malloc (cnlen + SNONCE_ENTROPY_BYTES + 1);
+	  state->sf.nonce = malloc (cnlen + snlen + 1);
 	  if (!state->sf.nonce)
 	    return GSASL_MALLOC_ERROR;
 
 	  memcpy (state->sf.nonce, state->cf.client_nonce, cnlen);
-	  memcpy (state->sf.nonce + cnlen, state->snonce,
-		  SNONCE_ENTROPY_BYTES);
-	  state->sf.nonce[cnlen + SNONCE_ENTROPY_BYTES] = '\0';
+	  memcpy (state->sf.nonce + cnlen, state->snonce, snlen);
+	  state->sf.nonce[cnlen + snlen] = '\0';
 	}
 
 	gsasl_property_set (sctx, GSASL_AUTHID, state->cf.username);
@@ -228,10 +276,23 @@ _gsasl_scram_sha1_server_step (Gsasl_session * sctx,
 
 	{
 	  const char *p = gsasl_property_get (sctx, GSASL_SCRAM_ITER);
+
 	  if (p)
 	    state->sf.iter = strtoul (p, NULL, 10);
 	  if (!p || state->sf.iter == 0 || state->sf.iter == ULONG_MAX)
 	    state->sf.iter = 4096;
+
+	  /* Save salt/iter as properties, so that client callback can
+	     access them. */
+	  {
+	    char *str = NULL;
+	    int n;
+	    n = asprintf (&str, "%zu", state->sf.iter);
+	    if (n < 0 || str == NULL)
+	      return GSASL_MALLOC_ERROR;
+	    gsasl_property_set (sctx, GSASL_SCRAM_ITER, str);
+	    free (str);
+	  }
 	}
 
 	{
@@ -241,6 +302,8 @@ _gsasl_scram_sha1_server_step (Gsasl_session * sctx,
 	      free (state->sf.salt);
 	      state->sf.salt = strdup (p);
 	    }
+	  else
+	    gsasl_property_set (sctx, GSASL_SCRAM_SALT, state->sf.salt);
 	}
 
 	rc = scram_print_server_first (&state->sf, &state->sf_str);
@@ -302,7 +365,7 @@ _gsasl_scram_sha1_server_step (Gsasl_session * sctx,
 	}
 
 	/* Base64 decode client proof and check that length matches
-	   SHA-1 size. */
+	   hash size. */
 	{
 	  size_t len;
 
@@ -310,65 +373,71 @@ _gsasl_scram_sha1_server_step (Gsasl_session * sctx,
 				  &state->clientproof, &len);
 	  if (rc != 0)
 	    return rc;
-	  if (len != 20)
+	  if (gsasl_hash_length (state->hash) != len)
 	    return GSASL_MECHANISM_PARSE_ERROR;
 	}
-
 	{
-	  const char *p;
+	  const char *p, *q;
 
 	  /* Get StoredKey and ServerKey */
-	  if ((p = gsasl_property_get (sctx, GSASL_PASSWORD)))
+	  if ((p = gsasl_property_get (sctx, GSASL_SCRAM_SERVERKEY))
+	      && (q = gsasl_property_get (sctx, GSASL_SCRAM_STOREDKEY)))
 	    {
-	      Gc_rc err;
-	      char *salt;
-	      size_t saltlen;
-	      char saltedpassword[20];
-	      char *clientkey;
-	      char *preppasswd;
-
-	      rc = gsasl_saslprep (p, 0, &preppasswd, NULL);
+	      rc = extract_serverkey (sctx, state, p, state->serverkey);
 	      if (rc != GSASL_OK)
 		return rc;
+	      rc = extract_serverkey (sctx, state, q, state->storedkey);
+	      if (rc != GSASL_OK)
+		return rc;
+	    }
+	  else if ((p = gsasl_property_get (sctx, GSASL_PASSWORD)))
+	    {
+	      char *salt;
+	      size_t saltlen;
+	      char saltedpassword[GSASL_HASH_MAX_SIZE];
+	      char hexstr[GSASL_HASH_MAX_SIZE * 2 + 1];
+	      char clientkey[GSASL_HASH_MAX_SIZE];
+	      char *b64str;
 
 	      rc = gsasl_base64_from (state->sf.salt, strlen (state->sf.salt),
 				      &salt, &saltlen);
+	      if (rc != GSASL_OK)
+		return rc;
+
+	      rc = gsasl_scram_secrets_from_password (state->hash,
+						      p,
+						      state->sf.iter,
+						      salt, saltlen,
+						      saltedpassword,
+						      clientkey,
+						      state->serverkey,
+						      state->storedkey);
+	      if (rc != GSASL_OK)
+		return rc;
+
+	      _gsasl_hex_encode (saltedpassword,
+				 gsasl_hash_length (state->hash), hexstr);
+	      gsasl_property_set (sctx, GSASL_SCRAM_SALTED_PASSWORD, hexstr);
+
+
+	      rc = gsasl_base64_to (state->serverkey,
+				    gsasl_hash_length (state->hash),
+				    &b64str, NULL);
 	      if (rc != 0)
-		{
-		  gsasl_free (preppasswd);
-		  return rc;
-		}
+		return rc;
+	      gsasl_property_set (sctx, GSASL_SCRAM_SERVERKEY, b64str);
+	      free (b64str);
+
+
+	      rc = gsasl_base64_to (state->storedkey,
+				    gsasl_hash_length (state->hash),
+				    &b64str, NULL);
+	      if (rc != 0)
+		return rc;
+	      gsasl_property_set (sctx, GSASL_SCRAM_STOREDKEY, b64str);
+	      free (b64str);
 
-	      /* SaltedPassword := Hi(password, salt) */
-	      err = gc_pbkdf2_sha1 (preppasswd, strlen (preppasswd),
-				    salt, saltlen,
-				    state->sf.iter, saltedpassword, 20);
-	      gsasl_free (preppasswd);
 	      gsasl_free (salt);
-	      if (err != GC_OK)
-		return GSASL_MALLOC_ERROR;
-
-	      /* ClientKey := HMAC(SaltedPassword, "Client Key") */
-#define CLIENT_KEY "Client Key"
-	      rc = gsasl_hmac_sha1 (saltedpassword, 20,
-				    CLIENT_KEY, strlen (CLIENT_KEY),
-				    &clientkey);
-	      if (rc != 0)
-		return rc;
-
-	      /* StoredKey := H(ClientKey) */
-	      rc = gsasl_sha1 (clientkey, 20, &state->storedkey);
-	      free (clientkey);
-	      if (rc != 0)
-		return rc;
-
-	      /* ServerKey := HMAC(SaltedPassword, "Server Key") */
-#define SERVER_KEY "Server Key"
-	      rc = gsasl_hmac_sha1 (saltedpassword, 20,
-				    SERVER_KEY, strlen (SERVER_KEY),
-				    &state->serverkey);
-	      if (rc != 0)
-		return rc;
 	    }
 	  else
 	    return GSASL_NO_PASSWORD;
@@ -394,46 +463,49 @@ _gsasl_scram_sha1_server_step (Gsasl_session * sctx,
 
 	  /* Check client proof. */
 	  {
-	    char *clientsignature;
-	    char *maybe_storedkey;
+	    char clientsignature[GSASL_HASH_MAX_SIZE];
+	    char maybe_storedkey[GSASL_HASH_MAX_SIZE];
 
 	    /* ClientSignature := HMAC(StoredKey, AuthMessage) */
-	    rc = gsasl_hmac_sha1 (state->storedkey, 20,
-				  state->authmessage,
-				  strlen (state->authmessage),
-				  &clientsignature);
+	    rc = _gsasl_hmac (state->hash,
+			      state->storedkey,
+			      gsasl_hash_length (state->hash),
+			      state->authmessage, strlen (state->authmessage),
+			      clientsignature);
 	    if (rc != 0)
 	      return rc;
 
 	    /* ClientKey := ClientProof XOR ClientSignature */
-	    memxor (clientsignature, state->clientproof, 20);
+	    memxor (clientsignature, state->clientproof,
+		    gsasl_hash_length (state->hash));
 
-	    rc = gsasl_sha1 (clientsignature, 20, &maybe_storedkey);
-	    free (clientsignature);
+	    rc = _gsasl_hash (state->hash, clientsignature,
+			      gsasl_hash_length (state->hash),
+			      maybe_storedkey);
 	    if (rc != 0)
 	      return rc;
 
-	    rc = memcmp (state->storedkey, maybe_storedkey, 20);
-	    free (maybe_storedkey);
+	    rc = memcmp (state->storedkey, maybe_storedkey,
+			 gsasl_hash_length (state->hash));
 	    if (rc != 0)
 	      return GSASL_AUTHENTICATION_ERROR;
 	  }
 
 	  /* Generate server verifier. */
 	  {
-	    char *serversignature;
+	    char serversignature[GSASL_HASH_MAX_SIZE];
 
 	    /* ServerSignature := HMAC(ServerKey, AuthMessage) */
-	    rc = gsasl_hmac_sha1 (state->serverkey, 20,
-				  state->authmessage,
-				  strlen (state->authmessage),
-				  &serversignature);
+	    rc = _gsasl_hmac (state->hash, state->serverkey,
+			      gsasl_hash_length (state->hash),
+			      state->authmessage,
+			      strlen (state->authmessage), serversignature);
 	    if (rc != 0)
 	      return rc;
 
-	    rc = gsasl_base64_to (serversignature, 20,
+	    rc = gsasl_base64_to (serversignature,
+				  gsasl_hash_length (state->hash),
 				  &state->sl.verifier, NULL);
-	    free (serversignature);
 	    if (rc != 0)
 	      return rc;
 	  }
@@ -457,7 +529,7 @@ _gsasl_scram_sha1_server_step (Gsasl_session * sctx,
 }
 
 void
-_gsasl_scram_sha1_server_finish (Gsasl_session * sctx, void *mech_data)
+_gsasl_scram_server_finish (Gsasl_session * sctx, void *mech_data)
 {
   struct scram_server_state *state = mech_data;
 
@@ -470,8 +542,6 @@ _gsasl_scram_sha1_server_finish (Gsasl_session * sctx, void *mech_data)
   free (state->sf_str);
   free (state->snonce);
   free (state->clientproof);
-  free (state->storedkey);
-  free (state->serverkey);
   free (state->authmessage);
   free (state->cbtlsunique);
   scram_free_client_first (&state->cf);
